function [psi,phi,freqs] = cauchy_wavelets(N,p,nb_wav,freqs_in)
% Calculates the Fourier transform of Cauchy wavelets

% Inputs :
% N : length of the signal (must be even)
% p : order of the wavelets
% nb_wav : number of wavelets (optional)
% freqs_in : characteristic frequencies of the wavelets (optional,
% overrides nb_wav)

% Outputs :
% psi : wavelets (in the Fourier domain)
% phi : low frequency component
% freqs : characteristic frequencies of the wavelets

% If freqs_in is not specified, dyadic wavelets will be returned,
% such that the highest frequency wavelet attains its maximum at
% N/2 (in the Fourier domain).
% If nb_wav is not specified either, the number of wavelets will be
% adjusted so that the lowest frequency wavelet attains its maximum
% between 1 and 2

% The low-frequency is a gaussian function :
%    phi(k) = exp(-(k/s)^2)
% (s is the index for which the wavelet of lowest frequency attains its
% maximum)

ks = [0:N-1] ;

if (nargin<4)
    
    if (nargin<3) || (isempty(nb_wav))
        nb_wav = floor(log(N/2)/log(2)) + 1 ;
    end
    m = N/2 ;
    
    for j=0:nb_wav-1
        psi(j+1,:) = (2^j*ks/m).^p .* exp(p*(1-2^j*ks/m)) ;
        freqs(j+1) = m*2^(-j) ;
    end
    
else

    for j=1:numel(freqs_in)
        psi(j,:) = (ks/freqs_in(j)).^p .* exp(p*(1-ks/freqs_in(j))) ;
    end
    freqs = freqs_in ;
    
end

if (nargout>1)
    [oub,ind_lowest] = min(freqs) ;
    [oub,ind] = max(psi(ind_lowest,:)) ;
    phi = exp(-(ks/ind).^2) ;
end