function [f_rec_gs,obj1,obj2] = reconstruct ...
    (fpsi_noisy,fphi_noisy,p,ln,opts)

% Reconstructs a function from the modulus of its Cauchy wavelet
% transform
% Inputs :
% - fpsi_noisy,fphi_noisy : modulus of the wavelet transform and
%                           low-frequency component
% - p : order of the Cauchy wavelets
% - ln : proportion of noise over the modulus
% - opts : options

[nbf,N] = size(fpsi_noisy) ;
if (mod(N,2)==1)
    error(['reconstruct: there is something wrong ; the array ' ...
           'containing the modulus should have an even number of ' ...
           'columns.']) ;
else
    N = N/2 ; % size of the signal
end

% read options about the number of iterations
if (nargin<5)
    opts = struct([]) ;
end
Kit = read_options(opts,'Kit',3000) ;
Kit_gs_rough = read_options(opts,'Kit_gs_rough',200) ;
Kit_gs = read_options(opts,'Kit_gs',2000) ;

nb_print = read_options(opts,'nb_print',1000) ;

% parameters for the termination criteria (chosen empirically)
prec_wav = min(3*ln,0.01) ; prec_mod = min(3*ln,0.05) ;

[psi,phi,freqs_wav] = cauchy_wavelets(N,p,nbf) ; % calculate wavelets


%% reconstruction of f by the Gerchberg-Saxton algorithm
% the reconstructed f is probably far from the original one ; it
% will only be used as a reference to estimate the errors over the
% products
f_ref = gerchberg_saxton ...
        ([],fpsi_noisy,fphi_noisy,psi,phi,Kit_gs_rough) ;
[fpsi_ref,fphi_ref] = wavelet_transform(f_ref,psi,phi) ;
% add noise
[fpsi_noisy_ref,fphi_noisy_ref] = ...
    add_global_noise(abs(fpsi_ref),fphi_ref,ln) ;


%% estimation of the products from the moduli
% The algorithm will at first try to reconstruct the f star psi_j
% for specific non-integer values of j
% characteristic frequencies of these psi_j's
freqs_eval = [freqs_wav(1)*(3/2),freqs_wav*(3/4)] ;
% wavelets and dual wavelets for these frequencies
psi_ev = cauchy_wavelets(N,p,[],freqs_eval) ;
[phi_conj,psi_conj] = dual_wavelets(phi,psi_ev) ;

% wavelet transform of the reference function for these frequencies
fpsi_eval_ref = wavelet_transform(f_ref,psi_ev,phi) ;

% estimation of the products
for s=1:nbf
    % from the rough estimation f_ref, determine a correct
    % value of the parameter ind, needed for the estimation of the
    % product
    true_prod_ref(s,:) = fpsi_eval_ref(s,:) ...
        .*conj(fpsi_eval_ref(s+1,:)) ;
    ind = eval_prod_ref(abs(fpsi_noisy_ref(s,:)).^2, ...
                               p,freqs_wav(s)*(3/2), ...
                               freqs_wav(s), ...
                               true_prod_ref(s,:)) ;
    % estimation of the product
    products(s,:) = eval_prod(abs(fpsi_noisy(s,:)).^2, p, ...
                              freqs_wav(s)*(3/2), ...
                              freqs_wav(s), ind) ;
end
% estimation of the low frequency component
ind = low_frequencies_ref(fphi_noisy_ref, ...
                          fpsi_eval_ref(nbf+1,:),phi, ...
                          freqs_wav(end)*(3/4),p) ;
f_low = low_frequencies(fphi_noisy,phi, ...
                        freqs_wav(end)*(3/4),p,ind) ;


%% Rough estimation of the f star psi_j's from the products
% Rough estimation of the maximal value of |f star psi_(j)|
for k=1:nbf
    ref_max(k) = 2*max(abs(fpsi_eval_ref(k,:))) ;    
end

% estimation of the f star psi_j's, for j corresponding to the
% values in freqs_eval
est(nbf+1,:) = f_low ;
for k=nbf:(-1):1
    % At the points where the estimated value is too small,
    % rectification (to avoid division by something very close to
    % zero)
    est(k+1,:) = est(k+1,:)./min(1,abs(est(k+1,:)) ...
                                 *ref_max(k)./abs(products(k,:))) ;
    % division
    est(k,:) = products(k,:)./conj(est(k+1,:)) ;
end
% estimation of f star phi
fphi_est = fphi_noisy ;

% estimation of f by inverse wavelet transform of est
ff_rec = phi_conj.*fft(fphi_est) + sum(fft(est,[],2).*psi_conj,1) ;
f_rec = ifft(ff_rec(1:N)) ;


%% We now reconstruct the f star psi_j's and the f.

% At each step of the algorithm, we get a new estimation est. We
% also calculate an estimate est_phi of f star phi and an estimate
% f_rec of f.

% The estimation est yields the values est_prod for the
% products. We want the following properties to be satisfied:
% |est_prod-products| small
% |est(nbf+1,:)-f_low| small
% |est_phi-fphi_noisy| small
% [est,est_phi] = wavelet transform of f_rec

% The algorithm thus consists in a minimization (by gradient
% descent) of the following function :
%  norm((est_prod-products)*(normalization coeff),2)^2
%  + norm((est(nbf+1,:)-f_low)*(normalization coeff),2)^2
%  + norm((est_phi-fphi_noisy)*(normalization coeff),2)^2
%  + alpha * norm([est,est_phi]-wavelet_transform(f_rec),2)^2

% normalization coefficients
inv_max = repmat(1./max(abs([products;f_low]),[],2),[1,2*N]) ;
inv_max = min(inv_max,10*min(inv_max(:))) ;
inv_max_phi = min(1/max(abs(fphi_noisy)),sqrt(10*min(inv_max(:)))) ;

% alpha is chosen so that the three first terms of the objective
% function are expected to be of the same order as the fourth one
alpha = (norm([products;f_low].*inv_max,'fro')^2 + ...
         + norm(inv_max_phi*fphi_noisy,2)^2) / ...
        (norm(fpsi_eval_ref(:),2)^2 + norm(fphi_est,2)^2) ;

step = 1 ; % initial step of the gradient descent
k_it = 0 ;
bool_term = 0 ; % boolean associated to the termination criterion


fprintf('Debut de la descente de gradient\n') ;
% gradient descent loop
while (k_it<Kit) && ~bool_term
    
    k_it = k_it + 1 ;
    
    if (mod(k_it,nb_print)==0)
        fprintf('%d iterations\n',k_it) ; % display info
    end
    if (mod(k_it,500)==0)
        step = norm(est(:),2)/norm(grad_est(:),2) ;
        % tries a large value for the step, in order to
        % accelerate convergence
    end
        
    % current estimation of the products
    est_prod = est(1:nbf,:).*conj(est(2:nbf+1,:)) ;
    % current error over the products
    err_prod = [est_prod-products ; est(nbf+1,:)-f_low] ;
    % wavelet transform of the current estimation of f
    [wav_trans,phi_rec] = wavelet_transform(f_rec,psi_ev,phi) ;

    % Current value of the objective function
    obj1(k_it) = norm(abs(err_prod(:)).*inv_max(:),2)^2 + ... 
        norm(abs(fphi_est-fphi_noisy)*inv_max_phi,2)^2 ;
    obj2(k_it) = alpha * (norm(wav_trans(:)-est(:),2)^2 ...
                          + norm(phi_rec-fphi_est,2)^2) ; 
    obj(k_it) = obj1(k_it)+obj2(k_it) ;
    
    % check whether the objective function has decreased or
    % increased since the previous step
    if (k_it>1) && (obj(k_it)>=obj(k_it-1))
        % if it has increased, go back to the previous values
        obj(k_it) = obj(k_it-1) ;
        obj1(k_it) = obj1(k_it-1) ;
        obj2(k_it) = obj2(k_it-1) ;
        
        est = prev_est ;
        fphi_est = prev_fphi_est ;
        f_rec = prev_f_rec ;
        % re-estimate the products
        est_prod = est(1:nbf,:).*conj(est(2:nbf+1,:)) ;
        err_prod = [est_prod-products ; est(nbf+1,:)-f_low] ;        
        [wav_trans,phi_rec] = wavelet_transform(f_rec,psi_ev,phi) ;
        % divide step by 2
        step = step/2 ;
    else        
        prev_est = est ;
        prev_fphi_est = fphi_est ;
        prev_f_rec = f_rec ;
    end
        
    
    % gradient
    grad_phi = 2 * (fphi_est-fphi_noisy)./ ...
        (1e-13 + abs(fphi_est-fphi_noisy)) ...
        .* abs(fphi_est-fphi_noisy) ...
        * inv_max_phi^2 ...
        + 2 * alpha * (fphi_est-phi_rec) ;
    
    grad_est = 2 * alpha * (est-wav_trans) ;
    err_prod_norm = err_prod./(1e-13 + abs(err_prod)) ;
    grad_est(1:nbf,:) = grad_est(1:nbf,:) + ...
        2 * est(2:nbf+1,:).*err_prod_norm(1:nbf,:) ...
        .*abs(err_prod(1:nbf,:)) ...
        .* inv_max(1:nbf,:).^2 ;
    grad_est(2:nbf+1,:) = grad_est(2:nbf+1,:) + ...
        2 * est(1:nbf,:).*conj(err_prod_norm(1:nbf,:)) ...
        .* abs(err_prod(1:nbf,:)) ...
            .* inv_max(1:nbf,:).^2 ;
    grad_est(nbf+1,:) = grad_est(nbf+1,:) + ...
        2 * err_prod_norm(nbf+1,:) .* ...
        abs(err_prod(nbf+1,:)) ...
        .* inv_max(nbf+1,:,:).^2 ;
    
    grad_f = 2 * alpha * adj_wav_trans( ...
        [wav_trans-est;phi_rec-fphi_est], ...
        conj(psi_ev),conj(phi)) ;

    
    % descent
    est = est-step*grad_est ;
    fphi_est = fphi_est-step*grad_phi ;
    f_rec = f_rec-step*grad_f ;

    % checks whether the termination criterion is satisfied

    % current distance between the estimation and the set of
    % admissible wavelet transforms
    e_wav = sqrt((norm(wav_trans(:)-est(:))^2 + ...
                  norm(phi_rec-fphi_est)^2) / ...
                 (norm(wav_trans(:))^2  + norm(phi_rec,2)^2)) ;

    % distance between the products of the estimations and the true
    % products
    p2 = [products;f_low] ;
    e_mod = sqrt((norm(err_prod(:))^2 + ... 
                  norm(abs(fphi_est-fphi_noisy))^2) / ...
                 (norm(p2(:),2)^2+norm(fphi_noisy,2)^2)) ;

    bool_term = (e_wav<prec_wav) && (e_mod<prec_mod) ;
    
end

%% improve the reconstruction with Gerchberg-Saxton
fprintf('Gerchberg-Saxton ... ') ;
f_rec_gs = gerchberg_saxton ...
    (prev_f_rec,fpsi_noisy,fphi_noisy,psi,phi,Kit_gs) ;
fprintf('end of Gerchberg-Saxton\n') ;